from enum import Enum
from typing import List, Optional
from uuid import UUID
from pydantic import (
    BaseModel,
    Field,
    model_validator,
    ConfigDict,
)


class DhGroup(str, Enum):
    """DH group options for IPsec crypto profiles."""

    NO_PFS = "no-pfs"
    GROUP1 = "group1"
    GROUP2 = "group2"
    GROUP5 = "group5"
    GROUP14 = "group14"
    GROUP19 = "group19"
    GROUP20 = "group20"


class EspEncryption(str, Enum):
    """ESP encryption algorithm options."""

    DES = "des"
    TRIPLE_DES = "3des"
    AES_128_CBC = "aes-128-cbc"
    AES_192_CBC = "aes-192-cbc"
    AES_256_CBC = "aes-256-cbc"
    AES_128_GCM = "aes-128-gcm"
    AES_256_GCM = "aes-256-gcm"
    NULL = "null"


class AhAuthentication(str, Enum):
    """AH authentication algorithm options."""

    MD5 = "md5"
    SHA1 = "sha1"
    SHA256 = "sha256"
    SHA384 = "sha384"
    SHA512 = "sha512"


class EspAuthentication(str, Enum):
    """ESP authentication algorithm options."""

    MD5 = "md5"
    SHA1 = "sha1"
    SHA256 = "sha256"
    SHA384 = "sha384"
    SHA512 = "sha512"


class LifetimeSeconds(BaseModel):
    """Lifetime specified in seconds."""

    seconds: int = Field(
        ...,
        description="Specify lifetime in seconds",
        ge=180,
        le=65535,
    )


class LifetimeMinutes(BaseModel):
    """Lifetime specified in minutes."""

    minutes: int = Field(
        ...,
        description="Specify lifetime in minutes",
        ge=3,
        le=65535,
    )


class LifetimeHours(BaseModel):
    """Lifetime specified in hours."""

    hours: int = Field(
        ...,
        description="Specify lifetime in hours",
        ge=1,
        le=65535,
    )


class LifetimeDays(BaseModel):
    """Lifetime specified in days."""

    days: int = Field(
        ...,
        description="Specify lifetime in days",
        ge=1,
        le=365,
    )


class LifesizeKB(BaseModel):
    """Lifesize specified in kilobytes."""

    kb: int = Field(
        ...,
        description="Specify lifesize in kilobytes(KB)",
        ge=1,
        le=65535,
    )


class LifesizeMB(BaseModel):
    """Lifesize specified in megabytes."""

    mb: int = Field(
        ...,
        description="Specify lifesize in megabytes(MB)",
        ge=1,
        le=65535,
    )


class LifesizeGB(BaseModel):
    """Lifesize specified in gigabytes."""

    gb: int = Field(
        ...,
        description="Specify lifesize in gigabytes(GB)",
        ge=1,
        le=65535,
    )


class LifesizeTB(BaseModel):
    """Lifesize specified in terabytes."""

    tb: int = Field(
        ...,
        description="Specify lifesize in terabytes(TB)",
        ge=1,
        le=65535,
    )


class EspConfig(BaseModel):
    """ESP configuration for IPsec crypto profiles."""

    encryption: List[EspEncryption] = Field(
        ...,
        description="Encryption algorithm",
    )
    authentication: List[str] = Field(
        ...,
        description="Authentication algorithm",
    )

    @model_validator(mode="before")
    def convert_string_to_enum(cls, values):
        """Convert string authentication values to EspAuthentication enum if needed."""
        auth = values.get("authentication")
        if auth and isinstance(auth, list):
            # Keep strings as is - the API expects string values
            # This validator helps accept both string values and enum values in tests
            pass
        return values


class AhConfig(BaseModel):
    """AH configuration for IPsec crypto profiles."""

    authentication: List[AhAuthentication] = Field(
        ...,
        description="Authentication algorithm",
    )


class IPsecCryptoProfileBaseModel(BaseModel):
    """Base model for IPsec Crypto Profiles containing fields common to all operations."""

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )

    @model_validator(mode="before")
    @classmethod
    def process_lifetime_and_lifesize(cls, values):
        """Handle different formats for lifetime and lifesize fields."""
        if not isinstance(values, dict):
            return values

        # Handle lifetime from Pydantic model instances
        lifetime = values.get("lifetime")
        if lifetime is None:
            return values

        # Convert LifetimeSeconds etc. objects to dicts
        if isinstance(lifetime, LifetimeSeconds):
            values["lifetime"] = {"seconds": lifetime.seconds}
        elif isinstance(lifetime, LifetimeMinutes):
            values["lifetime"] = {"minutes": lifetime.minutes}
        elif isinstance(lifetime, LifetimeHours):
            values["lifetime"] = {"hours": lifetime.hours}
        elif isinstance(lifetime, LifetimeDays):
            values["lifetime"] = {"days": lifetime.days}

        # Handle lifesize from Pydantic model instances
        lifesize = values.get("lifesize")
        if lifesize is not None:
            if isinstance(lifesize, LifesizeKB):
                values["lifesize"] = {"kb": lifesize.kb}
            elif isinstance(lifesize, LifesizeMB):
                values["lifesize"] = {"mb": lifesize.mb}
            elif isinstance(lifesize, LifesizeGB):
                values["lifesize"] = {"gb": lifesize.gb}
            elif isinstance(lifesize, LifesizeTB):
                values["lifesize"] = {"tb": lifesize.tb}

        return values

    name: str = Field(
        ...,
        description="Alphanumeric string begin with letter: [0-9a-zA-Z._-]",
        pattern=r"^[0-9a-zA-Z._\-]+$",
        max_length=31,
    )
    dh_group: Optional[DhGroup] = Field(
        default=DhGroup.GROUP2,
        description="Phase-2 DH group (PFS DH group)",
    )
    # Instead of using direct Union, we'll handle this field in the validator
    lifetime: dict = Field(
        ...,
        description="Lifetime configuration",
    )
    lifesize: Optional[dict] = Field(
        None,
        description="Lifesize configuration",
    )
    esp: Optional[EspConfig] = Field(
        None,
        description="ESP configuration",
    )
    ah: Optional[AhConfig] = Field(
        None,
        description="AH configuration",
    )

    # Container fields
    folder: Optional[str] = Field(
        None,
        pattern=r"^[a-zA-Z\d\-_. ]+$",
        max_length=64,
        description="The folder in which the resource is defined",
    )
    snippet: Optional[str] = Field(
        None,
        pattern=r"^[a-zA-Z\d\-_. ]+$",
        max_length=64,
        description="The snippet in which the resource is defined",
    )
    device: Optional[str] = Field(
        None,
        pattern=r"^[a-zA-Z\d\-_. ]+$",
        max_length=64,
        description="The device in which the resource is defined",
    )

    @model_validator(mode="after")
    def validate_security_protocol(self) -> "IPsecCryptoProfileBaseModel":
        """Validate that exactly one security protocol (ESP or AH) is configured."""
        if self.esp is not None and self.ah is not None:
            raise ValueError("Only one security protocol (ESP or AH) can be configured at a time")

        if self.esp is None and self.ah is None:
            raise ValueError("At least one security protocol (ESP or AH) must be configured")

        return self


class IPsecCryptoProfileCreateModel(IPsecCryptoProfileBaseModel):
    """Model for creating new IPsec Crypto Profiles."""

    @model_validator(mode="after")
    def validate_container(self) -> "IPsecCryptoProfileCreateModel":
        """Validate that exactly one container field is provided."""
        container_fields = ["folder", "snippet", "device"]
        provided = [field for field in container_fields if getattr(self, field) is not None]
        if len(provided) != 1:
            raise ValueError("Exactly one of 'folder', 'snippet', or 'device' must be provided.")
        return self


class IPsecCryptoProfileUpdateModel(IPsecCryptoProfileBaseModel):
    """Model for updating existing IPsec Crypto Profiles."""

    id: UUID = Field(
        ...,
        description="The UUID of the IPsec crypto profile",
        examples=["123e4567-e89b-12d3-a456-426655440000"],
    )


class IPsecCryptoProfileResponseModel(IPsecCryptoProfileBaseModel):
    """Model for IPsec Crypto Profile responses."""

    id: UUID = Field(
        ...,
        description="The UUID of the IPsec crypto profile",
        examples=["123e4567-e89b-12d3-a456-426655440000"],
    )
